#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Unittests for the 'Calibration' class."""

from pathlib import Path

import numpy as np
import pytest

from pyxel.calibration import Calibration
from pyxel.calibration.util import list_to_slice, read_data, read_single_data
from pyxel.configuration import Configuration, load
from pyxel.detectors import CCD
from pyxel.pipelines import DetectionPipeline, Processor

try:
    import pygmo as pg

    WITH_PYGMO = True
except ImportError:
    WITH_PYGMO = False


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize(
    "yaml",
    [
        "tests/data/calibrate.yaml",
        "tests/data/calibrate_sade.yaml",
        "tests/data/calibrate_sga.yaml",
        "tests/data/calibrate_nlopt.yaml",
        "tests/data/calibrate_models.yaml",
        "tests/data/calibrate_custom_fitness.yaml",
        "tests/data/calibrate_fits.yaml",
    ],
)
def test_set_algo(yaml):
    """Test"""
    cfg = load(yaml)
    calibration = cfg.calibration
    obj = calibration.algorithm.get_algorithm()
    if isinstance(obj, pg.sade):
        pass
    elif isinstance(obj, pg.sga):
        pass
    elif isinstance(obj, pg.nlopt):
        pass
    else:
        raise ReferenceError


@pytest.mark.parametrize(
    "input_data",
    [
        [Path("tests/data/calibrate-data.npy")],
        [Path("tests/data/calibrate-data.txt")],
        [Path("tests/data/ascii_input_3.data")],
        [Path("tests/data/ascii_input_4.data")],
        pytest.param(
            Path("tests/data/calibrate-data.npy"),
            marks=pytest.mark.xfail(raises=TypeError),
        ),
        pytest.param(
            Path("tests/data/calibrate-data.txt"),
            marks=pytest.mark.xfail(raises=TypeError),
        ),
        pytest.param(
            Path("tests/data/ascii_input_1.data"),
            marks=pytest.mark.xfail(raises=TypeError),
        ),
        pytest.param(
            Path("tests/data/ascii_input_2.data"),
            marks=pytest.mark.xfail(raises=TypeError),
        ),
    ],
)
def test_read_data(input_data):
    """Test"""
    output = read_data(input_data)
    assert isinstance(output, list)


@pytest.mark.parametrize(
    "input_data",
    [
        # pytest.param(
        #     [Path("tests/data/calibrate-data.npy")],
        #     marks=pytest.mark.xfail(raises=TypeError),
        # ),
        # pytest.param(
        #     [Path("tests/data/calibrate-data.txt")],
        #     marks=pytest.mark.xfail(raises=TypeError),
        # ),
        # pytest.param(
        #     [Path("tests/data/ascii_input_3.data")],
        #     marks=pytest.mark.xfail(raises=TypeError),
        # ),
        # pytest.param(
        #     [Path("tests/data/ascii_input_4.data")],
        #     marks=pytest.mark.xfail(raises=TypeError),
        # ),
        Path("tests/data/calibrate-data.npy"),
        Path("tests/data/calibrate-data.txt"),
        Path("tests/data/ascii_input_1.data"),
        Path("tests/data/ascii_input_2.data"),
    ],
)
def test_read_single_data(input_data):
    output = read_single_data(input_data)

    assert isinstance(output, np.ndarray)


@pytest.mark.parametrize(
    "input_data",
    [
        [1.0, 3.0, 2.0, 5.0],
        [1, 3, 2, 5],
        [0, 0, 0, 0],
        [2, 4, 1, 3, 2, 5],
        [4.0, 8.0, 1.0, 3.0, 2.0, 5.0],
        None,
        # [1],              # TODO
        # [1, 2, 3],
        # [1, 2, 3, 4, 5],
    ],
)
def test_list_to_slice(input_data):
    """Test"""
    output = list_to_slice(input_data)
    if isinstance(output, slice):
        pass
    elif isinstance(output, tuple) and all(isinstance(item, slice) for item in output):
        pass
    else:
        raise TypeError
    print(output)


@pytest.mark.skip(reason="!! FIX THIS TEST !!")
@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize("yaml", ["tests/data/calibrate_models.yaml"])
def test_run_calibration(yaml):
    """Test"""
    cfg = load(yaml)
    assert isinstance(cfg, Configuration)

    detector = cfg.ccd_detector
    assert isinstance(detector, CCD)

    pipeline = cfg.pipeline
    assert isinstance(pipeline, DetectionPipeline)

    processor = Processor(detector, pipeline)
    assert isinstance(processor, Processor)

    calibration = cfg.calibration
    assert isinstance(calibration, Calibration)

    assert calibration is not None
    result = calibration.run_calibration(processor)
    # assert result == 1         # TODO



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""TBW."""
import typing as t
from pathlib import Path

import dask.delayed as delayed
import pandas as pd
import xarray as xr
from dask.delayed import Delayed
from typing_extensions import Literal

from .outputs import Outputs

if t.TYPE_CHECKING:

    class SaveToFile(t.Protocol):
        """TBW."""

        def __call__(
            self, data: t.Any, name: str, with_auto_suffix: bool = True
        ) -> Path:
            """TBW."""
            ...


# Define type aliases
ValidName = Literal[
    "detector.image.array", "detector.signal.array", "detector.pixel.array"
]
ValidFormat = Literal["fits", "hdf", "npy", "txt", "csv", "png"]


class CalibrationOutputs(Outputs):
    """TBW."""

    def __init__(
        self,
        output_folder: t.Union[str, Path],
        save_data_to_file: t.Optional[
            t.Sequence[t.Mapping[ValidName, t.Sequence[ValidFormat]]]
        ] = None,
        save_calibration_data: t.Optional[
            t.Sequence[t.Mapping[str, t.Sequence[str]]]
        ] = None,
    ):
        super().__init__(
            output_folder=output_folder, save_data_to_file=save_data_to_file
        )

        # Parameter(s) specific for 'Calibration'
        self.save_calibration_data = (
            save_calibration_data
        )  # type: t.Optional[t.Sequence[t.Mapping[str, t.Sequence[str]]]]

    def save_processors(self, processors: pd.DataFrame) -> t.Sequence[Delayed]:
        """TBW."""
        lst = []  # type: t.List[delayed.Delayed]

        if self.save_data_to_file:

            for _, serie in processors.iterrows():
                id_island = serie["island"]  # type: int
                id_processor = serie["id_processor"]  # type: int
                processor = serie["processor"]  # type: Delayed

                # TODO: Create folders ?
                prefix = f"island{id_island:02d}_processor{id_processor:02d}"

                output_filenames = delayed(self.save_to_file)(
                    processor=processor, prefix=prefix, with_auto_suffix=False
                )  # type: Delayed
                lst.append(output_filenames)

        return lst

    def save_calibration_outputs(self, dataset: xr.Dataset, logs: pd.DataFrame) -> None:
        """Save the calibration outputs such as dataset and logs.

        Parameters
        ----------
        dataset: xr.Dataset
        logs: pd.DataFrame

        Returns
        -------
        None
        """

        save_methods = {"nc": self.save_to_netcdf}  # type: t.Dict[str, SaveToFile]

        if self.save_calibration_data is not None:

            for (
                dct
            ) in self.save_calibration_data:  # type: t.Mapping[str, t.Sequence[str]]
                first_item, *_ = dct.items()
                obj, format_list = first_item

                if obj == "logs":
                    for format in format_list:
                        if format == "csv":
                            filename = self.output_dir.joinpath("logs.csv")
                            logs.to_csv(filename)
                        elif format == "xlsx":
                            filename = self.output_dir.joinpath("logs.xlsx")
                            logs.to_excel(filename)
                        else:
                            raise NotImplementedError(
                                f"Saving to format {format} not implemented"
                            )

                elif obj == "dataset":

                    if format_list is not None:
                        for out_format in format_list:

                            if out_format not in save_methods.keys():
                                raise ValueError(
                                    "Format " + out_format + " not a valid save method!"
                                )

                            func = save_methods[out_format]
                            func(data=dataset, name=obj)

                else:
                    raise NotImplementedError(f"Object {obj} unknown.")



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel Charge class to generate electrons or holes inside detector."""

import typing as t

import numpy as np
import pandas as pd
from astropy.units import cds

from pyxel.data_structure import Particle

cds.enable()


class Charge(Particle):
    """Charged particle class defining and storing information of all electrons and holes.

    Properties stored are: charge, position, velocity, energy.
    """

    def __init__(self):
        # TODO: The following line is not really needed
        super().__init__()
        self.nextid = 0  # type: int

        self.columns = (
            "charge",
            "number",
            "init_energy",
            "energy",
            "init_pos_ver",
            "init_pos_hor",
            "init_pos_z",
            "position_ver",
            "position_hor",
            "position_z",
            "velocity_ver",
            "velocity_hor",
            "velocity_z",
        )  # type: t.Tuple[str, ...]

        self.EMPTY_FRAME = pd.DataFrame(
            columns=self.columns, dtype=float
        )  # type: pd.DataFrame

        self.frame = self.EMPTY_FRAME.copy()  # type: pd.DataFrame

    @staticmethod
    def create_charges(
        *,
        particle_type: str,  # TODO: Use Enum
        particles_per_cluster: np.ndarray,
        init_energy: np.ndarray,
        init_ver_position: np.ndarray,
        init_hor_position: np.ndarray,
        init_z_position: np.ndarray,
        init_ver_velocity: np.ndarray,
        init_hor_velocity: np.ndarray,
        init_z_velocity: np.ndarray,
    ) -> pd.DataFrame:
        """Create new charge(s) or group of charge(s) as a `DataFrame`.

        Parameters
        ----------
        particle_type : str
            Type of particle. Valid values: 'e' for an electron or 'h' for a hole.
        particles_per_cluster : array-like
        init_energy : array-like
        init_ver_position : array-like
        init_hor_position : array-like
        init_z_position : array-like
        init_ver_velocity : array-like
        init_hor_velocity : array-like
        init_z_velocity : array-like

        Returns
        -------
        dataframe
            Charge(s) stored in a `DataFrame`.
        """
        if not (
            len(particles_per_cluster)
            == len(init_energy)
            == len(init_ver_position)
            == len(init_hor_position)
            == len(init_z_position)
            == len(init_ver_velocity)
            == len(init_hor_velocity)
            == len(init_z_velocity)
        ):
            raise ValueError("List arguments have different lengths.")

        if not (
            particles_per_cluster.ndim
            == init_energy.ndim
            == init_ver_position.ndim
            == init_hor_position.ndim
            == init_z_position.ndim
            == init_ver_velocity.ndim
            == init_hor_velocity.ndim
            == init_z_velocity.ndim
            == 1
        ):
            raise ValueError("List arguments must have only one dimension.")

        elements = len(init_energy)

        # check_position(self.detector, init_ver_position, init_hor_position, init_z_position)      # TODO
        # check_energy(init_energy)             # TODO
        # Check if particle number is integer:
        # check_type(particles_per_cluster)      # TODO

        # TODO: particle_type should be a Enum class ?
        if particle_type == "e":
            charge = [-1] * elements  # * cds.e
        elif particle_type == "h":
            charge = [+1] * elements  # * cds.e
        else:
            raise ValueError("Given charged particle type can not be simulated")

        # if all(init_ver_velocity) == 0 and all(init_hor_velocity) == 0 and all(init_z_velocity) == 0:
        #     random_direction(1.0)

        # Create new charges as a `dict`
        new_charges = {
            "charge": charge,
            "number": particles_per_cluster,
            "init_energy": init_energy,
            "energy": init_energy,
            "init_pos_ver": init_ver_position,
            "init_pos_hor": init_hor_position,
            "init_pos_z": init_z_position,
            "position_ver": init_ver_position,
            "position_hor": init_hor_position,
            "position_z": init_z_position,
            "velocity_ver": init_ver_velocity,
            "velocity_hor": init_hor_velocity,
            "velocity_z": init_z_velocity,
        }  # type: t.Mapping[str, t.Union[t.Sequence, np.ndarray]]

        return pd.DataFrame(new_charges)

    def add_charge_dataframe(self, new_charges: pd.DataFrame) -> None:
        """Add new charge(s) or group of charge(s) inside the detector.

        Parameters
        ----------
        new_charges : DataFrame
            Charges as a `DataFrame`
        """
        if set(new_charges.columns) != set(self.columns):
            expected_columns = ", ".join(map(repr, self.columns))  # type: str
            raise ValueError(f"Expected columns: {expected_columns}")

        if self.frame.empty:
            new_frame = new_charges  # type: pd.DataFrame
        else:
            new_frame = self.frame.append(new_charges, ignore_index=True)

        self.frame = new_frame
        self.nextid = self.nextid + len(new_charges)

    def add_charge(
        self,
        *,
        particle_type: str,  # TODO: Use Enum
        particles_per_cluster: np.ndarray,
        init_energy: np.ndarray,
        init_ver_position: np.ndarray,
        init_hor_position: np.ndarray,
        init_z_position: np.ndarray,
        init_ver_velocity: np.ndarray,
        init_hor_velocity: np.ndarray,
        init_z_velocity: np.ndarray,
    ) -> None:
        """Add new charge(s) or group of charge(s) inside the detector.

        Parameters
        ----------
        particle_type : str
            Type of particle. Valid values: 'e' for an electron or 'h' for a hole.
        particles_per_cluster : array-like
        init_energy : array-like
        init_ver_position : array-like
        init_hor_position : array-like
        init_z_position : array-like
        init_ver_velocity : array-like
        init_hor_velocity : array-like
        init_z_velocity : array-like
        """
        # Create charge(s)
        new_charges = Charge.create_charges(
            particle_type=particle_type,
            particles_per_cluster=particles_per_cluster,
            init_energy=init_energy,
            init_ver_position=init_ver_position,
            init_hor_position=init_hor_position,
            init_z_position=init_z_position,
            init_ver_velocity=init_ver_velocity,
            init_hor_velocity=init_hor_velocity,
            init_z_velocity=init_z_velocity,
        )  # type: pd.DataFrame

        # Add charge(s)
        self.add_charge_dataframe(new_charges=new_charges)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel Array class."""

import typing as t  # noqa: F401

import numpy as np
from astropy.units import cds

from pyxel.util.memory import get_size

# TODO: Is it possible to move this to `data_structure/__init__.py' ?
cds.enable()


# TODO: Does it make sense to force 'self._array' to be read-only ?
#       It could be done with:
#       ... self._array = np.array(value)
#       ... self._array.setflags(write=False)
class Array:
    """Array class."""

    EXP_TYPE = type(None)  # type: t.Type
    TYPE_LIST = ()  # type: t.Tuple[t.Type, ...]

    # TODO: Add units ?
    def __init__(self, value: np.ndarray):
        self.validate_type(value)

        self._array = value  # type: np.ndarray

        self._numbytes = 0

        # TODO: is `self.type` needed ?
        # self.type = None            # type: t.Optional[type]

        # TODO: Implement a method to initialized 'self._array' ???

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        shape = self._array.shape
        dtype = self._array.dtype

        return f"{cls_name}<shape={shape}, dtype={dtype}>"

    def validate_type(self, value: np.ndarray) -> None:
        """Validate a value.

        Parameters
        ----------
        value
        """
        cls_name = self.__class__.__name__  # type: str

        if not isinstance(value, np.ndarray):
            raise TypeError(f"{cls_name} array should be a numpy.ndarray")

        if value.dtype not in self.TYPE_LIST:
            exp_type_name = self.EXP_TYPE.__name__  # type: str
            raise TypeError(f"Expected type of {cls_name} array is {exp_type_name}.")

    def validate_shape(self, value: np.ndarray) -> None:
        """TBW."""
        cls_name = self.__class__.__name__  # type: str

        if value.shape != self._array.shape:
            raise ValueError(f"Expected {cls_name} array is {self._array.shape}.")

    @property
    def array(self) -> np.ndarray:
        """Two dimensional numpy array storing the data.

        Only accepts an array with the right type and shape.
        """
        # if self._array is None:
        #     raise ValueError("'array' is not initialized.")

        return self._array

    @array.setter
    def array(self, value: np.ndarray) -> None:
        """Overwrite the two dimensional numpy array storing the data.

        Only accepts an array with the right type and shape.
        """
        self.validate_type(value)
        self.validate_shape(value)

        # self.type = value.dtype
        self._array = value

    # TODO: Is it necessary ? Maybe not if you implement method __array__
    @property
    def mean(self) -> float:
        """Return mean of all pixel values."""
        return float(np.mean(self._array))

    @property
    def std_deviation(self) -> float:
        """Return standard deviation of all pixel values."""
        return float(np.std(self._array))

    @property
    def max(self) -> float:
        """Return maximum of all pixel values."""
        return float(np.max(self._array))

    @property
    def min(self) -> float:
        """Return minimum of all pixel values."""
        return float(np.min(self._array))

    @property
    def peak_to_peak(self) -> float:
        """Return peak-to-peak value of all pixel values."""
        return float(np.ptp(self._array))

    @property
    def sum(self) -> float:
        """Return sum of all pixel values."""
        return float(np.sum(self._array))

    @property
    def numbytes(self) -> int:
        """Recursively calculates object size in bytes using Pympler library.

        Returns
        -------
        int
            Size of the object in bytes.
        """
        self._numbytes = get_size(self)
        return self._numbytes



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""Tools for display of calibration IO in notebooks."""

import os
import typing as t

import holoviews as hv
import numpy as np
import pandas as pd
import xarray as xr
from bokeh.models import PrintfTickFormatter

from pyxel.inputs import load_image

if t.TYPE_CHECKING:
    from pyxel.calibration import Calibration
    from pyxel.detectors import Detector


def display_calibration_inputs(
    calibration: "Calibration", detector: "Detector"
) -> hv.Layout:
    """Display calibration inputs and target data based on configuration file.

    Parameters
    ----------
    calibration: Calibration
        Instance of Calibration.
    detector: Detector
        Instance of Detector.

    Returns
    -------
    plot: hv.Layout
    """

    fnames_input = calibration.result_input_arguments[0].values
    fnames_target = [str(os.path.relpath(x)) for x in calibration.target_data_path]

    input_range = calibration.result_fit_range
    target_range = calibration.target_fit_range

    input_columns = input_range[3] - input_range[2]
    target_columns = target_range[3] - target_range[2]

    def get_data_input(data_id: int) -> t.Union[hv.Image, hv.Curve]:
        """Get input data based on configuration file.

        Parameters
        ----------
        data_id

        Returns
        -------
        im
        """

        # TODO: Fix typing, what if input arguments in calibration not strings?
        data = load_image(fnames_input[data_id])  # type: ignore
        if data.ndim == 1:
            im = hv.Curve((range(len(data)), data)).opts(
                tools=["hover"], aspect=1.5, xlabel="x", ylabel="z"
            )
        elif input_columns == 1:
            im = hv.Curve(
                (range(len(data[:, input_range[2]])), data[:, input_range[2]])
            ).opts(tools=["hover"], aspect=1.5, xlabel="x", ylabel="z")
        else:
            im = hv.Image((range(data.shape[1]), range(data.shape[0]), data)).opts(
                colorbar=True,
                cmap="gray",
                tools=["hover"],
                aspect=(detector.geometry.col / detector.geometry.row),
            )

        return im

    def get_data_target(data_id: int) -> t.Union[hv.Image, hv.Curve]:
        """Get target data based on the configuration file.

        Parameters
        ----------
        data_id

        Returns
        -------
        im
        """

        data = load_image(fnames_target[data_id])

        if data.ndim == 1:
            im = hv.Curve((range(len(data)), data)).opts(
                tools=["hover"], aspect=1.5, xlabel="x", ylabel="z"
            )
        elif target_columns == 1:
            im = hv.Curve(
                (range(len(data[:, target_range[2]])), data[:, target_range[2]])
            ).opts(tools=["hover"], aspect=1.5, xlabel="x", ylabel="z")
        else:
            im = hv.Image((range(data.shape[1]), range(data.shape[0]), data)).opts(
                colorbar=True,
                cmap="gray",
                tools=["hover"],
                aspect=(detector.geometry.col / detector.geometry.row),
            )

        return im

    ids = range(len(calibration.target_data_path))

    dmap1 = (
        hv.DynamicMap(get_data_input, kdims=["id"])
        .redim.values(id=ids)
        .relabel("Input")
    )
    dmap2 = (
        hv.DynamicMap(get_data_target, kdims=["id"])
        .redim.values(id=ids)
        .relabel("Target")
    )
    table = (
        hv.Table({"Input": fnames_input, "Target": fnames_target}, ["Input", "Target"])
        .opts(width=500)
        .relabel("Filenames")
    )

    plot = (
        dmap1.opts(framewise=True, axiswise=True)
        + dmap2.opts(framewise=True, axiswise=True)
        + table
    ).opts(tabs=True)

    return plot


def display_simulated(ds: xr.Dataset) -> hv.Layout:
    """Display simulated and target data from the output dataset.

    Parameters
    ----------
    ds: xr.Dataset
        Result dataset.

    Returns
    -------
    hv.Layout
    """
    result_type = ds.attrs["result_type"]

    var_name = {
        "ResultType.Image": "simulated_image",
        "ResultType.Signal": "simulated_signal",
        "ResultType.Pixel": "simulated_pixel",
    }[result_type]

    simulated_data = ds[var_name]
    target_data = ds["target"]

    output_data = xr.Dataset()
    output_data["simulated"] = simulated_data
    output_data["target"] = target_data
    output_data["residuals"] = target_data - simulated_data

    if len(output_data["target"].x) == 1:
        ds_target = hv.Dataset(output_data["target"].sel(x=0))
        ds_simulated = hv.Dataset(output_data["simulated"].sel(x=0))
        ds_residuals = hv.Dataset(output_data["residuals"].sel(x=0))

        plot_target_scatter = ds_target.to(hv.Scatter, kdims=["y"], dynamic=True)
        plot_target_line = ds_target.to(hv.Curve, kdims=["y"], dynamic=True)
        plot_simulated = ds_simulated.to(hv.Curve, kdims=["y"], dynamic=True)
        plot_residuals = ds_residuals.to(hv.Curve, kdims=["y"], dynamic=True)

        plot_target_scatter = plot_target_scatter.opts(
            logy=True,
            height=450,
            width=600,
            marker="o",
            size=5,
            fill_color=None,
            title="Target",
        )
        plot_target_line = plot_target_line.opts(
            logy=True,
            height=450,
            width=600,
            ylabel="Signal",
            title="Target",
            framewise=True,
            axiswise=True,
        )
        plot_simulated = plot_simulated.opts(
            logy=True,
            height=450,
            width=600,
            ylabel="Signal",
            title="Simulated",
            color="black",
            framewise=True,
            axiswise=True,
        )
        plot_residuals = plot_residuals.opts(
            height=450, width=600, title="Residuals", framewise=True, axiswise=True
        )

        overlay = (plot_simulated * plot_target_scatter).opts(title="Overlay")

        return (plot_simulated + plot_target_line + overlay + plot_residuals).opts(
            tabs=True
        )

    elif len(output_data["target"].y) == 1:
        ds_target = hv.Dataset(output_data["target"].sel(y=0))
        ds_simulated = hv.Dataset(output_data["simulated"].sel(y=0))
        ds_residuals = hv.Dataset(output_data["residuals"].sel(y=0))

        plot_target_scatter = ds_target.to(hv.Scatter, kdims=["x"], dynamic=True)
        plot_target_line = ds_target.to(hv.Curve, kdims=["x"], dynamic=True)
        plot_simulated = ds_simulated.to(hv.Curve, kdims=["x"], dynamic=True)
        plot_residuals = ds_residuals.to(hv.Curve, kdims=["x"], dynamic=True)

        plot_target_scatter = plot_target_scatter.opts(
            logy=True,
            height=450,
            width=600,
            marker="o",
            size=5,
            fill_color=None,
            title="Target",
        )
        plot_target_line = plot_target_line.opts(
            logy=True,
            height=450,
            width=600,
            ylabel="Signal",
            title="Target",
            framewise=True,
            axiswise=True,
        )
        plot_simulated = plot_simulated.opts(
            logy=True,
            height=450,
            width=600,
            ylabel="Signal",
            title="Simulated",
            color="black",
            framewise=True,
            axiswise=True,
        )
        plot_residuals = plot_residuals.opts(
            height=450, width=600, title="Residuals", framewise=True, axiswise=True
        )

        overlay = (plot_simulated.opts(color="black") * plot_target_scatter).opts(
            title="Overlay"
        )

        return (plot_simulated + plot_target_line + overlay + plot_residuals).opts(
            tabs=True
        )

    else:
        ds_target = hv.Dataset(output_data["target"])
        ds_simulated = hv.Dataset(output_data["simulated"])
        ds_residuals = hv.Dataset(output_data["residuals"])

        plot_target = ds_target.to(hv.Image, dynamic=True).relabel(label="Target")
        plot_simulated = ds_simulated.to(hv.Image, dynamic=True).relabel(
            label="Simulated"
        )
        plot_residuals = ds_residuals.to(hv.Image, dynamic=True).opts(title="Residuals")

        aspect = len(output_data["target"].y) / len(output_data["target"].x)

        plot_target = plot_target.opts(
            colorbar=True,
            cmap="gray",
            tools=["hover"],
            aspect=aspect,
            framewise=True,
            axiswise=True,
        )
        plot_simulated = plot_simulated.opts(
            colorbar=True,
            cmap="gray",
            tools=["hover"],
            aspect=aspect,
            framewise=True,
            axiswise=True,
        )
        plot_residuals = plot_residuals.opts(
            colorbar=True,
            cmap="gray",
            tools=["hover"],
            aspect=aspect,
            framewise=True,
            axiswise=True,
        )

        return (plot_simulated + plot_target + plot_residuals).opts(tabs=True)


def display_evolution(ds: xr.Dataset) -> hv.Layout:
    """Display best champion parameter and overall fitness vs evolution.

    Parameters
    ----------
    ds: xr.Dataset
        Result dataset.

    Returns
    -------
    plot: hv.Layout
        Output plot.
    """
    output_champions = xr.Dataset()
    output_champions["fitness"] = ds["champion_fitness"].drop(labels="evolution")  # type: ignore
    output_champions["parameters"] = ds["champion_parameters"].assign_coords(
        {"param_id": range(len(ds.param_id))}
    )

    ds_parameters = hv.Dataset(output_champions["parameters"])
    plot_parameters = ds_parameters.to(
        hv.Scatter, kdims=["evolution"], dynamic=True
    ).opts(aspect=1.5, axiswise=True, framewise=True, ylabel="Champion parameter")
    ds_fitness = hv.Dataset(output_champions["fitness"])
    plot_fitness = ds_fitness.to(hv.Scatter, kdims=["evolution"], dynamic=True).opts(
        aspect=1.5, axiswise=True, framewise=True, ylabel="Fitness"
    )

    plot = (
        plot_fitness.relabel("Fitness")
        + plot_parameters.relabel("Best champion parameters")
    ).opts(tabs=True)

    return plot


def optimal_parameters(ds: xr.Dataset) -> pd.DataFrame:
    """Return a dataframe of best parameters.

    Parameters
    ----------
    ds: xr.Dataset
        Result dataset.

    Returns
    -------
    best: pd.DataFrame
        Best champion parameters
    """

    best = (
        ds.champion_parameters.isel(evolution=-1)
        .isel(island=ds.champion_fitness.isel(evolution=-1).argmin())
        .to_dataframe()
    )
    return best


def champion_heatmap(
    ds: xr.Dataset,
    num_bins: int = 100,
    logx: bool = False,
    parameter_range: t.Optional[list] = None,
    island_range: t.Optional[list] = None,
    ind_range: t.Optional[list] = None,
) -> hv.Points:
    """Plot a heatmap of champion parameters vs fitness.

    Parameters
    ----------
    ds: xr.Dataset
        Result dataset.
    num_bins
        Number of bins, default is 100.
    logx: bool
        Logarithmic x axis.
    parameter_slice: slice
        Parameters slice.
    island_slice: slice
        Islands slice.
    ind_slice: slice
        Individuals slice.

    Returns
    -------
    plot: hv.Points
        Champion heatmap.
    """

    if parameter_range:
        parameter_slice = slice(parameter_range[0], parameter_range[1])
    else:
        parameter_slice = slice(None)
    if island_range:
        island_slice = slice(island_range[0], island_range[1])
    else:
        island_slice = slice(None)
    if ind_range:
        ind_slice = slice(ind_range[0], ind_range[1])
    else:
        ind_slice = slice(None)

    if "best_fitness" in ds:
        individuals = xr.Dataset()
        individuals["fitness"] = ds["best_fitness"].drop(labels="evolution")  # type: ignore
        individuals["parameters"] = ds["best_parameters"].assign_coords(
            {"param_id": range(len(ds.param_id))}
        )

        ind_id = individuals.coords["individual"].values

    output_champions = xr.Dataset()
    output_champions["fitness"] = ds["champion_fitness"].drop(labels="evolution")  # type: ignore
    output_champions["parameters"] = ds["champion_parameters"].assign_coords(
        {"param_id": range(len(ds.param_id))}
    )

    x = np.array([])
    y = np.array([])

    for parameter in output_champions.param_id[parameter_slice]:
        for island in output_champions.island[island_slice]:
            x = np.append(
                x,
                np.asarray(
                    output_champions.sel(param_id=parameter).sel(island=island)[
                        "parameters"
                    ]
                ),
            )
            y = np.append(
                y,
                np.asarray(
                    output_champions.sel(param_id=parameter).sel(island=island)[
                        "fitness"
                    ]
                ),
            )

            if "best_fitness" in ds:
                for k in ind_id[ind_slice]:
                    x = np.append(
                        x,
                        np.asarray(
                            individuals.sel(param_id=parameter)
                            .sel(island=island)
                            .sel(individual=k)["parameters"]
                        ),
                    )
                    y = np.append(
                        y,
                        np.asarray(
                            individuals.sel(param_id=parameter)
                            .sel(island=island)
                            .sel(individual=k)["fitness"]
                        ),
                    )

    # x=x[y<18500]
    # y=y[y<18500]

    # y=y[x>1e-7]
    # x=x[x>1e-7]

    if logx is True:
        bins = [
            np.geomspace(0.9 * np.min(x), 1.1 * np.max(x), num_bins + 1),
            num_bins,
        ]  # type: t.Union[list, int]
    else:
        bins = num_bins

    hist, xs, ys = np.histogram2d(x, y, bins=bins)

    df = pd.DataFrame(hist).stack().rename_axis(["x", "y"]).reset_index(name="val")
    df["x"] = [xs[x] for x in df["x"]]
    df["y"] = [ys[y] for y in df["y"]]
    df = df[(df[["val"]] != 0).all(axis=1)]
    df["val"] = np.array(df["val"], dtype=int)

    formatter = PrintfTickFormatter(format="%.0e")

    plot = hv.Points(df, kdims=["x", "y"], vdims=["val"]).opts(logx=logx, logz=True)
    plot = plot.opts(
        marker="s",
        height=650,
        width=900,
        tools=["hover"],
        color="val",
        colorbar=True,
        cmap="kbc_r",
        ylabel="Fitness",
        size=0.5,
        framewise=True,
        axiswise=True,
        colorbar_opts={"formatter": formatter},
        clabel="Occurence",
    )

    return plot



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
"""Pyxel examples downloader."""

import shutil
from os import listdir, remove, rmdir
from os.path import isdir, join
from zipfile import ZipFile

import requests
from tqdm import tqdm


def download_examples(foldername: str = "pyxel-examples", force: bool = False) -> None:
    """Download and save examples from Pyxel Data Gitlab repository in the working directory.

    Parameters
    ----------
    foldername: str
    force: bool

    Returns
    -------
    None
    """

    if isdir(foldername) and not force:
        raise OSError(
            f"Folder {foldername} already exists. Either delete it, use a different name, or use the force argument."
        )
    elif isdir(foldername) and force:
        shutil.rmtree(foldername)

    url = "https://gitlab.com/esa/pyxel-data/-/archive/master/pyxel-data-master.zip"
    response = requests.get(url, stream=True)

    total = int(response.headers.get("content-length", 0))
    with open("examples_tmp.zip", "wb") as tmp, tqdm(
        desc="Downloading examples",
        total=total,
        unit="B",
        unit_scale=True,
        unit_divisor=1024,
    ) as bar:
        for data in response.iter_content(chunk_size=1024):
            size = tmp.write(data)
            bar.update(size)

    with ZipFile("examples_tmp.zip", "r") as zipobj:
        zipobj.extractall(foldername)

    root = foldername
    for filename in listdir(join(root, "pyxel-data-master")):
        shutil.move(join(root, "pyxel-data-master", filename), join(root, filename))
    rmdir(join(root, "pyxel-data-master"))

    remove("examples_tmp.zip")

    print("Done.")



"""Unittests for the 'PixelLoader' class."""
#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import inspect
import re

import numpy as np
import pytest
import yaml

try:
    from yaml import UnsafeLoader as Loader  # For pyyaml >= 5.1
except ImportError:
    from yaml import Loader  # For pyyaml < 5.1


def _expr_processor(loader: Loader, node: yaml.ScalarNode):
    value = loader.construct_scalar(node)

    try:
        result = eval(value, {}, np.__dict__)

        if callable(result) or inspect.ismodule(result):
            result = value

    except NameError:
        result = value

    return result


Loader.add_implicit_resolver("!expr", re.compile(r"^.*$"), None)
Loader.add_constructor("!expr", _expr_processor)


@pytest.mark.parametrize(
    "data, expected",
    [
        ("a: !expr 3.14", {"a": 3.14}),
        ("x: !expr 2**16", {"x": 65536}),
        ("- !expr 3 * (4 + 5)", [27]),
        ("- !expr 1.0e-6", [0.000001]),
        ("- !expr 1.1E+2", [110.0]),
        ("- !expr 1.1e+2", [110.0]),
        ("- !expr 1e+2", [100.0]),
        ("a: !expr 2 * foo", {"a": "2 * foo"}),
        ("b: !expr pi * 2", {"b": 6.283185307179586}),
        ("- !expr sin(pi / 2)", [1.0]),
        ("- !expr range(3)", [range(3)]),
        ("- !expr range(3, 10, 2)", [range(3, 10, 2)]),
        ("!expr arange(4)", np.array([0, 1, 2, 3])),
        ("!expr random", "random"),
        ("random", "random"),
        ("a: 3.14", {"a": 3.14}),
        ("x: 2**16", {"x": 65536}),
        ("- 3 * (4 + 5)", [27]),
        ("- 1.0e-6", [0.000001]),
        ("- 1.1E+2", [110.0]),
        ("- 1.1e+2", [110.0]),
        ("- 1e+2", [100.0]),
        ("a: 2 * foo", {"a": "2 * foo"}),
        ("b: pi * 2", {"b": 6.283185307179586}),
        ("- sin(pi / 2)", [1.0]),
        ("- range(3)", [range(3)]),
        ("- range(3, 10, 2)", [range(3, 10, 2)]),
        ("arange(4)", np.array([0, 1, 2, 3])),
    ],
)
def test_expr_with_tag(data, expected):
    """Test tag '!expr'."""
    obj = yaml.load(data, Loader=Loader)

    if isinstance(expected, np.ndarray):
        # Test `numpy.ndarray`
        np.testing.assert_array_equal(obj, expected)
    else:
        # Normal test
        assert obj == expected



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel TARS model to generate charge by ionization."""

import bisect
import typing as t
from pathlib import Path

import numpy as np
import pandas as pd
from scipy import interpolate


def sampling_distribution(distribution: np.ndarray) -> float:
    """TBW.

    :param distribution:
    """
    u = np.random.random()  # type: float
    # random_value_from_dist = distribution[bisect.bisect(distribution[:, 1], u) - 1, 0]
    random_value_from_dist = get_xvalue_with_interpolation(
        function_array=distribution, y_value=u
    )

    return random_value_from_dist


def get_xvalue_with_interpolation(function_array: np.ndarray, y_value: float) -> float:
    """TBW.

    :param function_array:
    :param y_value:
    """
    if y_value <= function_array[0, 1]:
        intpol_x_value = function_array[0, 0]  # type: float
    elif y_value >= function_array[-1, 1]:
        intpol_x_value = function_array[-1, 0]
    else:
        y_index_bot = bisect.bisect(function_array[:, 1], y_value) - 1  # type: int
        y_index_top = y_index_bot + 1  # type: int
        y_value_bot = function_array[y_index_bot, 1]  # type: float
        y_value_top = function_array[y_index_top, 1]  # type: float
        x_value_bot = function_array[y_index_bot, 0]  # type: float
        x_value_top = function_array[y_index_top, 0]  # type: float

        intpol_x_value = x_value_bot + (y_value - y_value_bot) * (
            x_value_top - x_value_bot
        ) / (y_value_top - y_value_bot)

    return intpol_x_value


def get_yvalue_with_interpolation(function_array, x_value):
    """TBW.

    :param function_array:
    :param x_value:
    """
    x_index_bot = bisect.bisect(function_array[:, 0], x_value) - 1
    x_index_top = x_index_bot + 1
    x_value_bot = function_array[x_index_bot, 0]
    x_value_top = function_array[x_index_top, 0]
    y_value_bot = function_array[x_index_bot, 1]
    y_value_top = function_array[x_index_top, 1]

    intpol_y_value = y_value_bot + (x_value - x_value_bot) * (
        y_value_top - y_value_bot
    ) / (x_value_top - x_value_bot)

    return intpol_y_value


def load_histogram_data(
    file_name: Path,
    hist_type: str,
    skip_rows: t.Optional[int] = None,
    read_rows: t.Optional[int] = None,
) -> pd.DataFrame:
    """TBW.

    :param file_name:
    :param hist_type:
    :param skip_rows:
    :param read_rows:
    """
    # TODO store count in pandas dataframe as int !!!

    step_size_data = pd.read_csv(
        file_name,
        delimiter="\t",
        names=[hist_type, "counts"],
        usecols=[1, 2],
        skiprows=skip_rows,
        nrows=read_rows,
    )
    return step_size_data


# def load_energy_data(file_name, step_rows):
#     """TBW.
#
#     :param file_name:
#     :param step_rows:
#     :return:
#     """
#     # 'proton_' + energy + '_' + thickness + '_1M.ascii'
#     spectrum_data = pd.read_csv(file_name, delimiter="\t",
#                                 names=["energy", "counts"], usecols=[1, 2], skiprows=step_rows + 8)
#
#     return spectrum_data


def read_data(file_name: Path) -> np.ndarray:
    """TBW.

    :param file_name:
    :return:
    """
    full_path = file_name.resolve()
    if not full_path.exists():
        raise FileNotFoundError(f"Cannot find file '{full_path}' !")

    data = np.loadtxt(full_path, "float", "#")  # type: np.ndarray
    return data


def interpolate_data(data: np.ndarray) -> t.Callable[[np.ndarray], np.ndarray]:
    """TBW.

    :param data:
    :return:
    """
    data_function = interpolate.interp1d(
        data[:, 0], data[:, 1], kind="linear"
    )  # type: t.Callable[[np.ndarray], np.ndarray]
    return data_function
